"""
This script takes a sitemap file generated by `spider.py` and
spins up a number of workers (threads/processes/greenlets) to
hit those URLs.

A few additional features would be helpful here like:
 - remembering stats about the URLs requested
   + min, max, and avg response times
   + min, max, and avg response sizes
 - failure rates for URLs
 - response code counts (e.g.: 95 2xx, 4 4xx, and 1 5xx)

This script runs forever (or until it's killed) checking
these URLs and periodically outputing stats.

"""
import collections
from multiprocessing import Process
import logging
import signal
import time
import timeit

import gevent
from gevent import pool
from gevent import monkey
import requests
import yaml


monkey.patch_socket()
requests_log = logging.getLogger("requests")


def sitemap_to_urls(sitemap_path):
    with open(sitemap_path, "r") as sitemap_file:
        sitemap = yaml.load(sitemap_file.read())

    urls = []

    def append_paths(parent_path, d, urls):
        for key in d:
            url_path = parent_path + (key,)
            url = "/".join(url_path)
            urls.append(url)
            append_paths(url_path, d[key], urls)
    append_paths((), sitemap, urls)

    return urls



def takeN(size, iterable):
    batch = []
    for n, i in enumerate(iterable):
        batch.append(i)
        if n % size == 0:
            yield batch
            batch = []
        yield batch


base = ""


# STATS!
START_TIME = time.time()
TOTAL_ELAPSED = time.time() - START_TIME
LAST_NOTIFIED = 0.0
NOTIFICATION_INTERVAL = 60.0
COUNTER = collections.Counter()


def get_url(url):
    global LAST_NOTIFIED, TOTAL_ELAPSED

    absolute_url = base + url
    logging.debug("getting: %s ... ", absolute_url)
    start = timeit.default_timer()

    try:
        resp = requests.get(absolute_url, timeout=30.0)
    except requests.exceptions.Timeout:
        stop = timeit.default_timer()
        elapsed = stop - start
        logging.debug("%s: Timed out (%s)", absolute_url, elapsed)
        key = "{}:TIMEOUT".format(absolute_url)
        COUNTER.update({key: 1})
    except requests.exceptions.SSLError:
        stop = timeit.default_timer()
        elapsed = stop - start
        logging.debug("%s: SSL Error (%s)", absolute_url, elapsed)
        key = "{}:SSL_ERROR".format(absolute_url)
        COUNTER.update({key: 1})
    except requests.exceptions.ConnectionError:
        stop = timeit.default_timer()
        elapsed = stop - start
        logging.debug("%s: Connection Error (%s)", absolute_url, elapsed)
        key = "{}:CONNECT_ERROR".format(absolute_url)
        COUNTER.update({key: 1})
    else:
        stop = timeit.default_timer()
        elapsed = stop - start
        logging.debug("%s: %s (%s) (%4f)", absolute_url, resp.status_code, resp.reason, elapsed)
        key = "{}:{}".format(absolute_url, resp.status_code)
        COUNTER.update({key: 1})
    finally:
        TOTAL_ELAPSED = time.time() - START_TIME
        if TOTAL_ELAPSED > LAST_NOTIFIED + NOTIFICATION_INTERVAL:
            LAST_NOTIFIED = TOTAL_ELAPSED
            logging.info("most common response stats:")
            for x in COUNTER.most_common(10):
                logging.info("\t%s", x)


def run(inflight, urls, ttl):
    if ttl > 0:
        signal.alarm(ttl)

    gevent.reinit()
    p = pool.Pool(inflight)
    while True:
        p.map(get_url, urls)
        p.join()


def main(arguments):
    global base

    base = arguments.base_url
    sitemap_path = arguments.sitemap_path
    inflight = arguments.inflight
    processes = arguments.processes
    debug = arguments.debug
    ttl = arguments.ttl

    if debug:
        logging.basicConfig(level=logging.DEBUG,
                            format="%(asctime)s - %(levelname)s - %(message)s")
    else:
        requests_log.setLevel(logging.WARNING)
        logging.basicConfig(level=logging.INFO,
                            format="%(asctime)s - %(levelname)s - %(message)s")

    logging.info("spinneret generating traffic - URL=%s N=%s P=%s",
                 base, inflight, processes)

    urls = sitemap_to_urls(sitemap_path)

    if processes == 1:
        run(inflight, urls, ttl)
    else:
        process_list = [Process(target=run, args=(inflight, urls, ttl)) for x in xrange(processes)]
        for proc in process_list:
            proc.start()
        for proc in process_list:
            proc.join()



import argparse

from spinneret import spider
from spinneret import traffic


base = "http://localhost"
sitemap = 'sitemap.yaml'
inflight = 10
processes = 1


parser = argparse.ArgumentParser(description='crawl or generate traffic')
subparsers = parser.add_subparsers()


parser_spider = subparsers.add_parser('spider', help="crawl the specified webpage")
parser_spider.set_defaults(func=spider.main)
parser_spider.add_argument('-b', '--base_url',
                    metavar='BASE', type=str, default=base,
                    help='the URL to crawl')
parser_spider.add_argument('-s', '--sitemap_path',
                    metavar='SITEMAP', type=str, default=sitemap,
                    help='the path to store the sitemap yaml document')
parser_spider.add_argument('-d', '--debug',
                    action='store_true',
                    help='turn on debugging')


parser_traffic = subparsers.add_parser('traffic', help="generate traffic")
parser_traffic.set_defaults(func=traffic.main)
parser_traffic.add_argument('-b', '--base_url',
                    metavar='BASE', type=str, default=base,
                    help='request pages from this URL')
parser_traffic.add_argument('-s', '--sitemap_path',
                    metavar='SITEMAP', type=str, default=sitemap,
                    help='the sitemap.yaml to use to generate traffic')
parser_traffic.add_argument('-i', '--inflight',
                    metavar='N', type=int, default=inflight,
                    help='the number of requests to have in flight at any time')
parser_traffic.add_argument('-p', '--processes',
                    metavar='P', type=int, default=processes,
                    help='the number of processes; each process has N requests in flight')
parser_traffic.add_argument('-t', '--ttl',
                    metavar='SECONDS', type=int, default=0,
                    help='generate load and stop after SECONDS seconds')
parser_traffic.add_argument('-d', '--debug',
                    action='store_true',
                    help='turn on debugging')



#!/usr/bin/env python

import os

from setuptools import find_packages
from setuptools import setup


with open("requirements.txt", "r") as requirements_file:
    requirements = [req.strip() for req in requirements_file.read().split("\n")]


ROOT = os.path.dirname(__file__)


def version():
    init = os.path.join(ROOT, "spinneret", "__init__.py")
    version = None
    for line in open(init, "r"):
        if line.startswith("__version__"):
            version = line.split("=")[-1].strip().replace('\"', '')
    assert version is not None, "Unable to determine version!"
    return version


def long_description():
    readme = os.path.join(ROOT, "README.rst")
    long_description = open(readme, "r").read()
    return long_description


setup(name='spinneret',
      author="Mike Steder",
      author_email="steder@gmail.com",
      version=version(),
      description='Quick web spider and traffic generator',
      packages=find_packages(),
      package_data={'': ['config/*']},
      scripts=["bin/spinneret",],
      test_suite='nose.collector',
      install_requires=requirements,
      classifiers=["Development Status :: 3 - Alpha",
                   "Programming Language :: Python",
                   "Programming Language :: Python :: 2",
                   "Programming Language :: Python :: 2.7",
                   "Topic :: Utilities",],
      url="https://github.com/steder/spinneret",
      long_description=long_description(),
)



"""
This script clicks all the links on a page to generate some beta traffic
"""

import httplib
import logging
import ssl
import urllib2

import mechanize
import yaml


def is_internal_url(url):
    if url.startswith("/"):
        return True
    return False


image_extensions = ('gif',
                    'jpg',
                    'png')
def is_image_url(url):
    for extension in image_extensions:
        if url.endswith(extension):
            return True
    return False


def urls_to_sitemap(urls):
    sitemap = {"": {}}
    for url in urls:
        segments = (segment for segment in url.split("/") if segment != '')
        root = sitemap[""]
        for segment in segments:
            if root.get(segment):
                root = root.get(segment)
            else:
                root[segment] = {}
                root = root[segment]
    return sitemap


def get_depth(url):
    depth = len([segment for segment in url.split("/") if segment])
    return depth


def main(arguments):
    root = arguments.base_url
    debug = arguments.debug

    if debug:
        logging.basicConfig(level=logging.DEBUG,
                            format="%(asctime)s - %(levelname)s - %(message)s")
    else:
        logging.basicConfig(level=logging.INFO,
                            format="%(asctime)s - %(levelname)s - %(message)s")

    logging.info("spinneret spidering - URL=%s", root)

    bro = mechanize.Browser()
    bro.open(root)

    test_bro = mechanize.Browser()

    visited = {}
    ignored = {}

    def visit_links(url, links, max_depth=3):
        logging.info("visiting links on: %s (nLinks=%s)", url, len(links))
        # for link in links:
        while links:
            link = links.pop(0)
            depth = get_depth(link.url)
            if (is_internal_url(link.url) and
                not is_image_url(link.url) and
                link.url not in visited and
                depth <= max_depth):
                visited[link.url] = link.absolute_url
                try:
                    logging.debug("Opening %s: ...", link.absolute_url)
                    test_bro.open(link.absolute_url)
                    if depth < max_depth:
                        try:
                            links.extend(link for link in test_bro.links()
                                         if link.url not in visited and link.url not in ignored)
                        except mechanize.BrowserStateError:
                            logging.error("Error getting links for: %s", link.absolute_url)
                except urllib2.HTTPError:
                    logging.error("Error opening: %s", link.absolute_url)
                except httplib.BadStatusLine:
                    logging.error("Bad status error: %s", link.absolute_url)
                else:
                    logging.info("%s OK", link.absolute_url)
            else:
                ignored[link.url] = link.absolute_url

    visit_links(root, list(bro.links()))

    sitemap = urls_to_sitemap(visited.keys())
    with open("sitemap.yaml", "w") as f:
        yaml.dump(sitemap, f)
    with open("links.yaml", "w") as f2:
        yaml.dump(visited, f2)
